import typing

from fntypes import Option

from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.dispatch.view.base import BaseView
from telegrinder.tools.functional import from_optional

Event = typing.TypeVar("Event", bound=BaseCute)
Data = typing.TypeVar("Data")

ECHO = lambda x: x


class Hasher(typing.Generic[Event, Data]):
    def __init__(
        self,
        view: type[BaseView[Event]],
        get_hash_from_data: typing.Callable[[Data], typing.Hashable | None] | None = None,
        get_data_from_event: typing.Callable[[Event], Data | None] | None = None,
    ):
        self.view = view
        self._get_hash_from_data = get_hash_from_data
        self._get_data_from_event = get_data_from_event

    def get_name(self) -> str:
        return f"{self.view.__class__.__name__}_{id(self)}"

    def get_hash_from_data(self, data: Data) -> Option[typing.Hashable]:
        if self._get_hash_from_data is None:
            raise NotImplementedError
        return from_optional(self._get_hash_from_data(data))

    def get_data_from_event(self, event: Event) -> Option[Data]:
        if not self._get_data_from_event:
            raise NotImplementedError
        return from_optional(self._get_data_from_event(event))

    def get_hash_from_data_from_event(self, event: Event) -> Option[typing.Hashable]:
        return self.get_data_from_event(event).and_then(self.get_hash_from_data)  # type: ignore

    def __hash__(self) -> int:
        return hash(self.get_name())

    def __repr__(self) -> str:
        return f"<Hasher {self.get_name()}>"


__all__ = ("Hasher",)



import datetime
import typing

from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.func import FuncHandler
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.bot.dispatch.waiter_machine.short_state import ShortStateContext
from telegrinder.modules import logger

from .hasher import Hasher

if typing.TYPE_CHECKING:
    from .machine import WaiterMachine
    from .short_state import ShortState

EventType = typing.TypeVar("EventType", bound=BaseCute)


class WaiterMiddleware(ABCMiddleware[EventType]):
    def __init__(
        self,
        machine: "WaiterMachine",
        hasher: Hasher,
    ) -> None:
        self.machine = machine
        self.hasher = hasher

    async def pre(self, event: EventType, ctx: Context) -> bool:
        if self.hasher not in self.machine.storage:
            return True

        key = self.hasher.get_hash_from_data_from_event(event)
        if key is None:
            logger.info(f"Unable to get hash from event with hasher {self.hasher}")
            return True

        short_state: "ShortState[EventType] | None" = self.machine.storage[self.hasher].get(key.unwrap())
        if not short_state:
            return True

        preset_context = Context(short_state=short_state)
        if short_state.context is not None:
            preset_context.update(short_state.context.context)

        # Run filter rule
        if short_state.filter and not await check_rule(
            event.ctx_api, short_state.filter, ctx.raw_update, preset_context
        ):
            logger.debug("Filter rule {!r} failed", short_state.filter)
            return True

        if short_state.expiration_date is not None and datetime.datetime.now() >= short_state.expiration_date:
            await self.machine.drop(
                self.hasher,
                self.hasher.get_data_from_event(event).unwrap(),
                **preset_context.copy(),
            )
            return True

        handler = FuncHandler(
            self.pass_runtime,
            [short_state.release] if short_state.release else [],
            dataclass=None,
            preset_context=preset_context,
        )
        result = await handler.check(event.ctx_api, ctx.raw_update, ctx)

        if result is True:
            await handler.run(event.api, event, ctx)

        elif on_miss := short_state.actions.get("on_miss"):  # noqa: SIM102
            if await on_miss.check(event.ctx_api, ctx.raw_update, ctx):
                await on_miss.run(event.ctx_api, event, ctx)

        return False

    async def pass_runtime(
        self,
        event: EventType,
        short_state: "ShortState[EventType]",
        ctx: Context,
    ) -> None:
        short_state.context = ShortStateContext(event, ctx)
        short_state.event.set()


__all__ = ("WaiterMiddleware",)



import dataclasses
import importlib
import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.node.base import ComposeError, Node
from telegrinder.node.update import UpdateNode

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.process import check_rule
    from telegrinder.bot.rules.abc import ABCRule


class RuleChain(dict[str, typing.Any], Node):
    dataclass = dict
    rules: tuple["ABCRule", ...] = ()

    def __init_subclass__(cls, *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init_subclass__(*args, **kwargs)

        if cls.__name__ == "_RuleNode":
            return
        cls.dataclass = cls.generate_node_dataclass(cls)

    def __new__(cls, *rules: "ABCRule") -> type[Node]:
        return type("_RuleNode", (cls,), {"dataclass": dict, "rules": rules})  # type: ignore

    def __class_getitem__(cls, items: "ABCRule | tuple[ABCRule, ...]", /) -> typing.Self:
        if not isinstance(items, tuple):
            items = (items,)
        return cls(*items)

    @staticmethod
    def generate_node_dataclass(cls_: type["RuleChain"]):  # noqa: ANN205
        return dataclasses.dataclass(type(cls_.__name__, (object,), dict(cls_.__dict__)))

    @classmethod
    async def compose(cls, update: UpdateNode) -> typing.Any:
        # Hack to avoid circular import
        globalns = globals()
        if "check_rule" not in globalns:
            globalns.update(
                {
                    "check_rule": getattr(
                        importlib.import_module("telegrinder.bot.dispatch.process"),
                        "check_rule",
                    ),
                },
            )

        ctx = Context()
        for rule in cls.rules:
            if not await check_rule(update.api, rule, update, ctx):
                raise ComposeError(f"Rule {rule!r} failed!")

        try:
            if dataclasses.is_dataclass(cls.dataclass):
                return cls.dataclass(**{k: ctx[k] for k in cls.__annotations__})
            return cls.dataclass(**ctx)
        except Exception as exc:
            raise ComposeError(f"Dataclass validation error: {exc}")

    @classmethod
    def as_node(cls) -> type[typing.Self]:
        return cls

    @classmethod
    def get_subnodes(cls) -> dict[typing.Literal["update"], type[UpdateNode]]:
        return {"update": UpdateNode}

    @classmethod
    def is_generator(cls) -> typing.Literal[False]:
        return False


__all__ = ("RuleChain",)



"""This is an implementation of GNU gettext (pyBabel)."""

import gettext
import os

from telegrinder.tools.i18n import ABCI18n
from telegrinder.tools.i18n.base import ABCTranslator


class SimpleI18n(ABCI18n):
    def __init__(self, folder: str, domain: str, default_locale: str):
        self.folder = folder
        self.domain = domain
        self.default_locale = default_locale
        self.translators = self._load_translators()

    def _load_translators(self) -> dict[str, gettext.GNUTranslations]:
        result = {}
        for name in os.listdir(self.folder):
            if not os.path.isdir(os.path.join(self.folder, name)):
                continue

            mo_path = os.path.join(self.folder, name, "LC_MESSAGES", f"{self.domain}.mo")
            if os.path.exists(mo_path):
                with open(mo_path, "rb") as f:
                    result[name] = gettext.GNUTranslations(f)
            elif os.path.exists(mo_path[:-2] + "po"):
                raise FileNotFoundError(".po files should be compiled first")
        return result

    def get_translator_by_locale(self, locale: str) -> "SimpleTranslator":
        return SimpleTranslator(locale, self.translators.get(locale, self.translators[self.default_locale]))


class SimpleTranslator(ABCTranslator):
    def __init__(self, locale: str, g: gettext.GNUTranslations):
        self.g = g
        super().__init__(locale)

    def get(self, __key: str, *args, **kwargs) -> str:
        return self.g.gettext(__key).format(*args, **kwargs)


__all__ = ("SimpleI18n", "SimpleTranslator")



import re
import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.node.text import Text

from .abc import ABCRule

PatternLike: typing.TypeAlias = str | typing.Pattern[str]


class Regex(ABCRule):
    def __init__(self, regexp: PatternLike | list[PatternLike]) -> None:
        self.regexp: list[re.Pattern[str]] = []
        match regexp:
            case re.Pattern() as pattern:
                self.regexp.append(pattern)
            case str(regex):
                self.regexp.append(re.compile(regex))
            case _:
                self.regexp.extend(
                    re.compile(regexp) if isinstance(regexp, str) else regexp for regexp in regexp
                )

    async def check(self, text: Text, ctx: Context) -> bool:
        for regexp in self.regexp:
            response = re.match(regexp, text)
            if response is not None:
                if matches := response.groupdict():
                    ctx |= matches
                else:
                    ctx |= {"matches": response.groups() or (response.group(),)}
                return True
        return False


__all__ = ("Regex",)



import typing
from abc import ABC, abstractmethod

from telegrinder.bot.cute_types.base import BaseCute

if typing.TYPE_CHECKING:
    from telegrinder.api import API
    from telegrinder.bot.dispatch.view.abc import ABCStateView

EventT = typing.TypeVar("EventT", bound=BaseCute)


class ABCScenario(ABC, typing.Generic[EventT]):
    @abstractmethod
    def wait(self, api: "API", view: "ABCStateView[EventT]") -> typing.Any:
        pass


__all__ = ("ABCScenario",)



from telegrinder.bot.cute_types import MessageCute as Message
from telegrinder.bot.dispatch.view import MessageView

from .hasher import Hasher


def from_chat_hash(chat_id: int) -> int:
    return chat_id


def get_chat_from_event(event: Message) -> int:
    return event.chat.id


MESSAGE_IN_CHAT = Hasher(
    view=MessageView, get_hash_from_data=from_chat_hash, get_data_from_event=get_chat_from_event
)


def from_user_hash(from_id: int) -> int:
    return from_id


def get_user_from_event(event: Message) -> int:
    return event.from_user.id


MESSAGE_FROM_USER = Hasher(
    view=MessageView,
    get_hash_from_data=from_user_hash,
    get_data_from_event=get_user_from_event,
)


def from_user_in_chat_hash(chat_and_user: tuple[int, int]) -> str:
    return f"{chat_and_user[0]}_{chat_and_user[1]}"


def get_user_in_chat_from_event(event: Message) -> tuple[int, int]:
    return event.chat.id, event.from_user.id


MESSAGE_FROM_USER_IN_CHAT = Hasher(
    view=MessageView,
    get_hash_from_data=from_user_in_chat_hash,
    get_data_from_event=get_user_in_chat_from_event,
)


__all__ = (
    "MESSAGE_FROM_USER",
    "MESSAGE_FROM_USER_IN_CHAT",
    "MESSAGE_IN_CHAT",
)



import typing

from fntypes.result import Error, Ok

from telegrinder.api.api import API
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.return_manager.abc import ABCReturnManager
from telegrinder.model import Model
from telegrinder.modules import logger
from telegrinder.node.composer import CONTEXT_STORE_NODES_KEY, NodeScope, compose_nodes
from telegrinder.tools.i18n.base import I18nEnum
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.handler.abc import ABCHandler
    from telegrinder.bot.rules.abc import ABCRule

Event = typing.TypeVar("Event", bound=Model)


async def process_inner(
    api: API,
    event: Event,
    raw_event: Update,
    middlewares: list[ABCMiddleware[Event]],
    handlers: list["ABCHandler[Event]"],
    return_manager: ABCReturnManager[Event] | None = None,
) -> bool:
    logger.debug("Processing {!r}...", event.__class__.__name__)
    ctx = Context(raw_update=raw_event)
    ctx[CONTEXT_STORE_NODES_KEY] = {}  # For per-event shared nodes

    logger.debug("Run pre middlewares...")
    for middleware in middlewares:
        middleware_result = await middleware.pre(event, ctx)
        logger.debug("Middleware {!r} returned: {!r}", middleware.__class__.__qualname__, middleware_result)
        if middleware_result is False:
            return False

    found = False
    responses = []
    ctx_copy = ctx.copy()

    for handler in handlers:
        if await handler.check(api, raw_event, ctx):
            logger.debug("Handler {!r} matched, run...", handler)
            found = True
            response = await handler.run(api, event, ctx)
            logger.debug("Handler {!r} returned: {!r}", handler, response)
            responses.append(response)
            if return_manager is not None:
                await return_manager.run(response, event, ctx)
            if handler.is_blocking:
                break

        ctx = ctx_copy

    logger.debug("Run post middlewares...")
    for middleware in middlewares:
        logger.debug("Run post middleware {!r}", middleware.__class__.__qualname__)
        await middleware.post(event, responses, ctx)

    for session in ctx.get(CONTEXT_STORE_NODES_KEY, {}).values():
        await session.close(scopes=(NodeScope.PER_EVENT,))

    logger.debug(
        "{} handlers, returns {!r}",
        "No found" if not found else "Found",
        found,
    )
    return found


async def check_rule(
    api: API,
    rule: "ABCRule",
    update: Update,
    ctx: Context,
) -> bool:
    """Checks requirements, adapts update.
    Returns check result."""

    # Running adapter
    match await rule.adapter.adapt(api, update, ctx):
        case Ok(value):
            adapted_value = value
        case Error(err):
            logger.debug("Adapter failed with error message: {!r}", str(err))
            return False

    # Preparing update
    if isinstance(adapted_val := ctx.get(rule.adapter.ADAPTED_VALUE_KEY or ""), UpdateCute):
        update = adapted_val
    elif isinstance(adapted_value, UpdateCute):
        update = adapted_value
    else:
        update = UpdateCute.from_update(update, bound_api=api)

    # Running subrules to fetch requirements
    ctx_copy = ctx.copy()
    for requirement in rule.requires:
        if not await check_rule(api, requirement, update, ctx_copy):
            return False

    # Translating translatable rules
    if I18nEnum.I18N in ctx:
        rule = await rule.translate(ctx[I18nEnum.I18N])

    ctx |= ctx_copy

    # Composing required nodes
    nodes = rule.required_nodes
    node_col = None
    if nodes:
        result = await compose_nodes(nodes, ctx, data={Update: update, API: api})
        if not result:
            return False
        node_col = result.value

    # Running check
    result = await rule.bounding_check(adapted_value, ctx, node_col=node_col)

    # Closing node sessions if there are any
    if node_col is not None:
        await node_col.close_all()

    return result


__all__ = ("check_rule", "process_inner")
